﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Higuchi.Net.Mail
{
	/// <summary>
	/// A manager to handle buffers for the socket connections
	/// </summary>
	/// <remarks>
	/// When used in an async call a buffer is pinned. Large numbers of pinned buffers
	/// cause problem with the GC (in particular it causes heap fragmentation).
	///
	/// This class maintains a set of large segments and gives clients pieces of these
	/// segments that they can use for their buffers. The alternative to this would be to
    /// create many small arrays which it then maintains. This methodology should be slightly
	/// better than the many small array methodology because in creating only a few very
	/// large Objects it will force these Objects to be placed on the LOH. Since the
	/// Objects are on the LOH they are at this time not subject to compacting which would
	/// require an update of all GC roots as would be the case with lots of smaller arrays
	/// that were in the normal heap.
	/// </remarks>
	public class BufferManager
	{
		private readonly Int32 _SegmentChunks;
		private readonly Int32 _ChunkSize;
		private readonly Int32 _SegmentSize;
		private readonly Stack<ArraySegment<Byte>> _Buffers;
		private readonly Object _LockObject = new Object();
		private readonly List<Byte[]> _Segments;
		/// <summary>
		/// The current number of buffers available
		/// </summary>
		public Int32 AvailableBuffers
		{
			get { return _Buffers.Count; } //do we really care about volatility here?
		}
		/// <summary>
		/// The total size of all buffers
		/// </summary>
		public Int32 TotalBufferSize
		{
			get { return _Segments.Count * _SegmentSize; } //do we really care about volatility here?
		}
		/// <summary>
		/// Creates a new segment, makes buffers available
		/// </summary>
		private void CreateNewSegment()
		{
			Byte[] Bytes = new Byte[_SegmentChunks * _ChunkSize];
			_Segments.Add(Bytes);
			for (Int32 i = 0; i < _SegmentChunks; i++)
			{
				ArraySegment<Byte> chunk = new ArraySegment<Byte>(Bytes, i * _ChunkSize, _ChunkSize);
				_Buffers.Push(chunk);
			}
		}
		/// <summary>
		/// Checks out a buffer from the manager
		/// </summary>
		/// <remarks>
		/// It is the client's responsibility to return the buffer to the manger by
		/// calling <see cref="CheckIn"></see> on the buffer
		/// </remarks>
		/// <returns>A ArraySegment that can be used as a buffer</returns>
		public ArraySegment<Byte> CheckOut()
		{
			lock (_LockObject)
			{
				if (_Buffers.Count == 0)
				{
					CreateNewSegment();
				}
				return _Buffers.Pop();
			}
		}
		/// <summary>
		/// Returns a buffer to the control of the manager
		/// </summary>
		/// <remarks>
		/// It is the client's responsibility to return the buffer to the manger by
		/// calling <see cref="CheckIn"></see> on the buffer
		/// </remarks>
		/// <param name="buffer">The ArraySegment to return to the cache</param>
		public void CheckIn(ArraySegment<Byte> buffer)
		{
			lock (_LockObject)
			{
				_Buffers.Push(buffer);
			}
		}
		/// <summary>
		/// Constructs a new <see cref="BufferManager"></see> Object
		/// </summary>
		/// <param name="segmentChunks">The number of chunks tocreate per segment</param>
		/// <param name="chunkSize">The size of a chunk in Bytes</param>
		public BufferManager(Int32 segmentChunks, Int32 chunkSize) :
			this(segmentChunks, chunkSize, 1) { }
		/// <summary>
		/// Constructs a new <see cref="BufferManager"></see> Object
		/// </summary>
		/// <param name="segmentChunks">The number of chunks tocreate per segment</param>
		/// <param name="chunkSize">The size of a chunk in Bytes</param>
		/// <param name="initialSegments">The initial number of segments to create</param>
		public BufferManager(Int32 segmentChunks, Int32 chunkSize, Int32 initialSegments)
		{
			_SegmentChunks = segmentChunks;
			_ChunkSize = chunkSize;
			_SegmentSize = _SegmentChunks * _ChunkSize;
			_Buffers = new Stack<ArraySegment<Byte>>(segmentChunks * initialSegments);
			_Segments = new List<Byte[]>();
			for (Int32 i = 0; i < initialSegments; i++)
			{
				this.CreateNewSegment();
			}
		}
	}
}