﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using Higuchi.Net.Mail;

namespace Higuchi.Net.Pop3
{
    /// Represent pop3 content.
    /// <summary>
    /// Represent pop3 content.
    /// </summary>
    public class Pop3Content : MimeContent
    {
        private Pop3Message _Message;
        private Pop3Content _ParentContent = null;
        private String _Data;
        /// デコード済みの本文の文字列データ
        /// <summary>
        /// Field for decoded body data.
        /// デコード済みの本文の文字列データ
        /// </summary>
        private String _BodyText;
        private Boolean _BodyTextCreated = false;
        private List<Pop3Content> _Contents = new List<Pop3Content>();
        /// 親のContentを取得または設定します。
        /// <summary>
        /// Get or set parent content object.
        /// 親のContentを取得または設定します。
        /// </summary>
        public Pop3Content ParentContent
        {
            get { return this._ParentContent; }
            private set { this._ParentContent = value; }
        }
        /// Nameの値を取得します。
        /// <summary>
        /// Get name value.
        /// Nameの値を取得します。
        /// </summary>
        public String Name
        {
            get { return this.ContentType.Name; }
        }
        /// FileNameの値を取得します。
        /// <summary>
        /// Get filename value.
        /// FileNameの値を取得します。
        /// </summary>
        public String FileName
        {
            get { return this.ContentDisposition.FileName; }
        }
        /// Body部の文字列を取得します。
        /// <summary>
        /// Get body text of this mail.
        /// Body部の文字列を取得します。
        /// </summary>
        public String BodyText
        {
            get
            {
                this.EnsureBodyText();
                return this._BodyText;
            }
            set { this._BodyText = value; }
        }
        /// このインスタンスを生成するときに使用した文字列データを取得します。
        /// <summary>
        /// Get text data used to create this instance.
        /// このインスタンスを生成するときに使用した文字列データを取得します。
        /// </summary>
        public String Data
        {
            get { return this._Data; }
        }
        /// Pop3Contentのコレクションを取得します。
        /// <summary>
        /// Get pop3 content collection of this mail.
        /// Pop3Contentのコレクションを取得します。
        /// </summary>
        public new List<Pop3Content> Contents
        {
            get { return this._Contents; }
        }
        /// Body部分のテキストが生成済みかどうかを示す値を取得します。
        /// <summary>
        /// Get value that indicate body text is created or not.
        /// Body部分のテキストが生成済みかどうかを示す値を取得します。
        /// </summary>
        protected Boolean BodyTextCreated
        {
            get { return this._BodyTextCreated; }
            set { this._BodyTextCreated = value; }
        }
		/// <summary>
		/// 
		/// </summary>
		/// <param name="message"></param>
		/// <param name="text"></param>
        public Pop3Content(Pop3Message message, String text) : 
            base(text)
        {
            this.Initialize(message, text);
        }
        /// 初期化処理を行います。
        /// <summary>
        /// 初期化処理を行います。
        /// </summary>
        /// <param name="message"></param>
        /// <param name="text"></param>
        private void Initialize(Pop3Message message, String text)
        {
            Pop3Content ct = null;

            this._Message = message;
            this._Contents = new List<Pop3Content>();
            this._Data = text;
            this._BodyText = "";
            if (this.IsMultiPart == true)
            {
                List<String> l=  MimeContent.ParseToContentTextList(this.BodyData, this.MultiPartBoundary);
                for (int i = 0; i < l.Count; i++)
                {
                    ct = new Pop3Content(this._Message, l[i]);
                    ct.ParentContent = this;
                    this._Contents.Add(ct);
                }
            }
        }
        /// Body部のテキストがセットされているか確認し、セットされてない場合はBody部の文字列をセットします。
        /// <summary>
        /// Body部のテキストがセットされているか確認し、セットされてない場合はBody部の文字列をセットします。
        /// </summary>
        /// <returns></returns>
        protected virtual void EnsureBodyText()
        {
            if (this.BodyTextCreated == false)
            {
                if (this.ContentType.Value.IndexOf("message/rfc822") > -1)
                {
                    this.BodyText = this.BodyData;
                }
                else if (this.IsText == true)
                {
                    this.BodyText = MailParser.DecodeFromMailBody(this.BodyData, this.ContentTransferEncoding, this.ContentEncoding);
                }
                else
                {
                    this.BodyText = this.BodyData;
                }
            }
            this.BodyTextCreated = true;
        }
        /// このインスタンスの値を元に、SmtpContentクラスのインスタンスを生成します。
        /// <summary>
        /// Create SmtpContent instance with this instance value.
        /// このインスタンスの値を元に、SmtpContentクラスのインスタンスを生成します。
        /// </summary>
        /// <returns></returns>
        public Smtp.SmtpContent CreateSmtpContent()
        {
            Smtp.SmtpContent ct = new Higuchi.Net.Smtp.SmtpContent();
            Field f = null;

            for (int i = 0; i < this.Header.Count; i++)
            {
                f = this.Header[i];
                if (String.IsNullOrEmpty(f.Value) == true)
                { continue; }
                ct[f.Key] = MailParser.DecodeFromMailHeaderLine(f.Value);
            }
            for (int i = 0; i < this.ContentType.Fields.Count; i++)
            {
                f = this.ContentType.Fields[i];
                ct.ContentType.Fields.Add(new Field(f.Key, MailParser.DecodeFromMailHeaderLine(f.Value)));
            }
            for (int i = 0; i < this.ContentDisposition.Fields.Count; i++)
            {
                f = this.ContentDisposition.Fields[i];
                ct.ContentDisposition.Fields.Add(new Field(f.Key, MailParser.DecodeFromMailHeaderLine(f.Value)));
            }
            ct.BodyText = this.BodyText;
            for (int i = 0; i < this.Contents.Count; i++)
            {
                ct.Contents.Add(this.Contents[i].CreateSmtpContent());
            }
            return ct;
        }
    }
}
