﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Net.Security;
using System.Net.NetworkInformation;
using System.Security.Cryptography.X509Certificates;
using Higuchi.Net.Mail;
using Higuchi.Net.Smtp.Command;
using Higuchi.Net.Pop3;

namespace Higuchi.Net.Smtp
{
    /// Represent and probide functionality about smtp command.
    /// <summary>
    /// Represent and probide functionality about smtp command.
    /// </summary>
    public class SmtpClient : IDisposable
    {
		/// <summary>
		/// Default smtp port
		/// </summary>
        public const Int32 DefaultPort = 25;
		/// <summary>
		/// Default ssl port
		/// </summary>
        public const Int32 DefaultSslPort = 443;
        private SmtpAuthenticateMode _Mode = SmtpAuthenticateMode.Auto;
        private String _UserName = "";
        private String _Password = "";
        private String _ServerName = "127.0.0.1";
        private Int32 _Port = SmtpClient.DefaultPort;
        private String _HostName = "";
        private Boolean _Ssl = false;
        private Boolean _Tls = false;
        private Int32 _ReceiveTimeout = 10 * 1000;
		private Int32 _SendBufferSize = 8192;
		private Int32 _ReceiveBufferSize = 8192;
		private Encoding _ResponseEncoding = Encoding.ASCII;
		private Boolean _PopBeforeSmtp = false;
        private Pop3Client _Pop3Client = new Pop3Client();
        private TcpClient _TcpClient = null;
        private Stream _Stream = null;
		private AutoResetEvent _SendCommandDone = new AutoResetEvent(false);
		private AutoResetEvent _GetResponseDone = new AutoResetEvent(false);
        private Int32 _ThreadSleepMilliseconds = 10;
        private SmtpConnectionState _State = SmtpConnectionState.Disconnected;
        private Boolean _Commnicating = false;
        /// 認証の方法を取得または設定します。
        /// <summary>
        /// 認証の方法を取得または設定します。
        /// </summary>
        public SmtpAuthenticateMode AuthenticateMode
        {
            get { return this._Mode; }
            set { this._Mode = value; }
        }
        /// 認証に使用するユーザー名を取得または設定します。
        /// <summary>
        /// 認証に使用するユーザー名を取得または設定します。
        /// </summary>
        public String UserName
        {
            get { return this._UserName; }
            set { this._UserName = value; }
        }
        /// 認証に使用するパスワードを取得または設定します。
        /// <summary>
        /// 認証に使用するパスワードを取得または設定します。
        /// </summary>
        public String Password
        {
            get { return this._Password; }
            set { this._Password = value; }
        }
        /// SMTPメールサーバーのサーバー名を取得または設定します。
        /// <summary>
        /// SMTPメールサーバーのサーバー名を取得または設定します。
        /// </summary>
        public String ServerName
        {
            get { return this._ServerName; }
            set { this._ServerName = value; }
        }
        /// 通信に使用するPort番号を取得または設定します。
        /// <summary>
        /// 通信に使用するPort番号を取得または設定します。
        /// </summary>
        public Int32 Port
        {
            get { return this._Port; }
            set { this._Port = value; }
        }
        /// 送信元マシンのホスト名を取得または設定します。
        /// <summary>
        /// 送信元マシンのホスト名を取得または設定します。
        /// </summary>
        public String HostName
        {
            get { return this._HostName; }
            set { this._HostName = value; }
        }
        /// 通信をSSLで暗号化するかどうかを示す値を取得または設定します。
        /// <summary>
        /// 通信をSSLで暗号化するかどうかを示す値を取得または設定します。
        /// </summary>
        public Boolean Ssl
        {
            get { return this._Ssl; }
            set { this._Ssl = value; }
        }
        /// 通信にTLSを使用するかどうかを示す値を取得します。
        /// <summary>
        /// 通信にTLSを使用するかどうかを示す値を取得します。
        /// </summary>
        public Boolean Tls
        {
            get { return this._Tls; }
            set { this._Tls = value; }
        }
        /// 受信処理のタイムアウトの秒数をミリ秒単位で取得または設定します。
        /// <summary>
        /// 受信処理のタイムアウトの秒数をミリ秒単位で取得または設定します。
        /// </summary>
        public Int32 ReceiveTimeout
        {
            get { return this._ReceiveTimeout; }
            set
            {
                this._ReceiveTimeout = value;
                if (this._TcpClient != null)
                {
                    this._TcpClient.ReceiveTimeout = this._ReceiveTimeout;
                }
            }
        }
		/// 送信データのバッファサイズを取得または設定します。
		/// <summary>
		/// Get or set buffer size to send.
		/// 送信データのバッファサイズを取得または設定します。
		/// </summary>
		public Int32 SendBufferSize
		{
			get { return this._SendBufferSize; }
			set
			{
				this._SendBufferSize = value;
				if (this._TcpClient != null)
				{
					this._TcpClient.SendBufferSize = this._SendBufferSize;
				}
			}
		}
		/// 受信データのバッファサイズを取得または設定します。
        /// <summary>
        /// 受信データのバッファサイズを取得または設定します。
        /// </summary>
        public Int32 ReceiveBufferSize
        {
            get { return this._ReceiveBufferSize; }
            set
            {
                this._ReceiveBufferSize = value;
                if (this._TcpClient != null)
                {
                    this._TcpClient.ReceiveBufferSize = this._ReceiveBufferSize;
                }
            }
        }
        /// 受信データのエンコーディングを取得または設定します。
        /// <summary>
        /// 受信データのエンコーディングを取得または設定します。
        /// </summary>
        public Encoding ResponseEncoding
        {
            get { return _ResponseEncoding; }
            set { _ResponseEncoding = value; }
        }
        /// 受信データを取得する際にサーバーからのレスポンスデータを待機する秒数をミリ秒単位で取得または設定します。
        /// <summary>
        /// 受信データを取得する際にサーバーからのレスポンスデータを待機する秒数をミリ秒単位で取得または設定します。
        /// </summary>
        public Int32 ThreadSleepMilliseconds
        {
            get { return this._ThreadSleepMilliseconds; }
            set { this._ThreadSleepMilliseconds = value; }
        }
        /// 接続の状態を示す値を取得します。
        /// <summary>
        /// 接続の状態を示す値を取得します。
        /// </summary>
        public SmtpConnectionState State
        {
            get { return this._State; }
        }
        /// サーバーへ接続済みかどうかを示す値を取得します。
        /// <summary>
        /// サーバーへ接続済みかどうかを示す値を取得します。
        /// </summary>
        public Boolean Available
        {
            get { return this._State != SmtpConnectionState.Disconnected; }
        }
        /// サーバーと通信中かどうかを示す値を取得します。
        /// <summary>
        /// サーバーと通信中かどうかを示す値を取得します。
        /// サーバーにCommandを送信してから受信データを全て受信し終わるまでの間、trueを返します。
        /// </summary>
        public Boolean Commnicating
        {
            get { return this._Commnicating; }
        }
        /// PopBeforeSmtp認証を行うかどうかを示す値を取得または設定します。
        /// <summary>
        /// PopBeforeSmtp認証を行うかどうかを示す値を取得または設定します。
        /// </summary>
        public Boolean PopBeforeSmtp
        {
            get { return this._PopBeforeSmtp; }
            set { this._PopBeforeSmtp = value; }
        }
        /// PopBeforeSmtp認証を行う場合に使用されるPop3Clientを取得します。
        /// <summary>
        /// PopBeforeSmtp認証を行う場合に使用されるPop3Clientを取得します。
        /// </summary>
        public Pop3Client Pop3Client
        {
            get { return this._Pop3Client; }
        }
		/// <summary>
		/// 
		/// </summary>
        public SmtpClient()
        {
        }
        /// 接続先のサーバーと通信を行うためのTcpClientオブジェクトを取得します。
        /// <summary>
        /// 接続先のサーバーと通信を行うためのTcpClientオブジェクトを取得します。
        /// </summary>
        /// <returns></returns>
        private TcpClient GetTcpClient()
        {
            TcpClient tc = null;
            IPHostEntry hostEntry = null;

			//サーバー名からIPアドレスのリストを取得します。
			hostEntry = this.GetHostEntry();
			//有効なIPアドレスかどうか判別し、有効なIPアドレスにセットされたソケットを取得します。
			if (hostEntry != null)
			{
				foreach (IPAddress address in hostEntry.AddressList)
				{
					tc = this.TryGetTcpClient(address);
					if (tc != null) { break; }
				}
			}
			if (tc == null)
			{
				this._State = SmtpConnectionState.Disconnected;
				return null;
			}
			this._State = SmtpConnectionState.Connected;
			return tc;
        }
		private IPHostEntry GetHostEntry()
		{
			try
			{
				return Dns.GetHostEntry(this._ServerName);
			}
			catch { }
			return null;
		}
		private TcpClient TryGetTcpClient(IPAddress address)
		{
			IPEndPoint ipe = new IPEndPoint(address, this._Port);
			TcpClient tc = null;

			try
			{
				tc = new TcpClient(ipe.AddressFamily);
				tc.Connect(ipe);
				if (tc.Connected == true)
				{
					tc.ReceiveTimeout = this._ReceiveTimeout;
					tc.SendBufferSize = this._SendBufferSize;
					tc.ReceiveBufferSize = this._ReceiveBufferSize;
				}
			}
			catch
			{
				tc = null;
			}
			return tc;
		}
		private static Boolean RemoteCertificateValidationCallback(Object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
        {
            return true;
        }
		/// サーバーへの接続を開きます。
		/// <summary>
		/// サーバーへの接続を開きます。
		/// </summary>
		public SmtpConnectionState Open()
		{
			SmtpCommandResult rs = null;

			this._TcpClient = this.GetTcpClient();
			if (this._TcpClient == null)
			{
				this._State = SmtpConnectionState.Disconnected;
			}
			else
			{
				if (this._Ssl == true)
				{
					SslStream ssl = new SslStream(this._TcpClient.GetStream(), true, SmtpClient.RemoteCertificateValidationCallback);
					ssl.AuthenticateAsClient(this._ServerName);
					if (ssl.IsAuthenticated == false)
					{
						this._State = SmtpConnectionState.Disconnected;
						this._TcpClient = null;
						return this._State;
					}
					this._Stream = ssl;
				}
				else
				{
					this._Stream = this._TcpClient.GetStream();
				}
				rs = this.GetResponse();
				if (rs.StatusCode == SmtpCommandResultCode.ServiceReady)
				{
					this._State = SmtpConnectionState.Connected;
				}
				else
				{
					this._TcpClient = null;
					this._State = SmtpConnectionState.Disconnected;
				}
			}
			return this._State;
		}
		/// サーバーへの接続が開かれていない場合、サーバーへの接続を開きます。
		/// <summary>
		/// サーバーへの接続が開かれていない場合、サーバーへの接続を開きます。
		/// </summary>
		public SmtpConnectionState EnsureOpen()
		{
			if (this._TcpClient != null)
			{ return this._State; }
			return this.Open();
		}
		private void SendCommand(String command)
        {
			this.SendCommand(Encoding.ASCII.GetBytes(command + MailParser.NewLine));
        }
		private void SendCommand(Byte[] bytes)
		{
			AsynchronousSendContext cx = null;

            if (this._TcpClient == null)
            {
                throw new SmtpConnectException("Smtp connection is closed");
            }
			try
			{
				cx = new AsynchronousSendContext(Encoding.ASCII, bytes);
				cx.FillBuffer();
				_Stream.BeginWrite(cx.Buffer.Array, 0, cx.SendBufferSize, this.SendCommandCallback, cx);
				_SendCommandDone.WaitOne();
            }
			catch (Exception ex)
			{
				throw new SmtpSendException(ex);
			}
			finally
			{
				if (cx != null)
				{
					cx.Dispose();
				}
			}
            //Throw exception that occor other thread.
            if (cx.Exception != null)
            {
                throw cx.Exception;
            }
        }
		private void SendCommandCallback(IAsyncResult result)
		{
            AsynchronousSendContext cx = null;
            try
            {
                cx = (AsynchronousSendContext)result.AsyncState;
                _Stream.EndWrite(result);
                if (cx.IsSendAllBytes == true)
                {
                    cx.FillBuffer();
                    _Stream.BeginWrite(cx.Buffer.Array, 0, cx.SendBufferSize, this.SendCommandCallback, cx);
                }
                else
                {
                    _SendCommandDone.Set();
                }
            }
            catch (Exception ex)
            {
                cx.Exception = ex;
            }
            if (cx.Exception != null)
            {
                try
                {
                    _SendCommandDone.Set();
                }
                catch (ObjectDisposedException) { }
            }
        }
		private SmtpCommandResult GetResponse()
        {
			SmtpCommandResultLine CurrentLine = null;
            List<SmtpCommandResultLine> l = new List<SmtpCommandResultLine>();
			StringReader sr  = null;

			if (this._TcpClient == null)
            {
                throw new SmtpConnectException("Connection to SMTP server is closed");
            }
			using (var cx = new AsynchronousSmtpResponseContext(_ResponseEncoding))
			{
				_Stream.BeginRead(cx.Buffer.Array, 0, cx.Buffer.Array.Length, this.GetResponseCallback, cx);
				var bl = _GetResponseDone.WaitOne(_ReceiveTimeout);
                if (cx.Exception != null)
                {
                    throw cx.Exception;
                }
                if (cx.Timeout == true || bl == false)
				{
					throw new SmtpReceiveException("Response timeout");
				}
				sr = new StringReader(_ResponseEncoding.GetString(cx.Data.ToArray()));
				while (true)
				{
					CurrentLine = new SmtpCommandResultLine(sr.ReadLine());
					if (CurrentLine.InvalidFormat == true)
					{
						throw new SmtpReceiveException("Response format is invalid");
					}
					l.Add(CurrentLine);
					//次の行があるかチェック
					if (CurrentLine.HasNextLine == false)
					{ break; }
				}
				this.SetSmtpCommandState();
			}
            return new SmtpCommandResult(l.ToArray());
        }
		private void GetResponseCallback(IAsyncResult result)
		{
            AsynchronousSmtpResponseContext cx = null;

            try
            {
                cx = (AsynchronousSmtpResponseContext)result.AsyncState;
                if (this._TcpClient == null)
                {
                    throw new SmtpConnectException("Connection to SMTP server is closed");
                }
                Int32 readBytes = _Stream.EndRead(result);
                TimeSpan ts = DateTime.Now - cx.StartTime;

                if (ts.TotalMilliseconds > this._ReceiveTimeout)
                {
                    cx.Timeout = true;
                    _GetResponseDone.Set();
                }
                if (cx.ReadBuffer(readBytes) == true)
                {
                    _Stream.BeginRead(cx.Buffer.Array, 0, cx.Buffer.Array.Length, this.GetResponseCallback, cx);
                }
                else
                {
                    _GetResponseDone.Set();
                }
            }
            catch (Exception ex)
            {
                cx.Exception = ex;
            }
            if (cx.Exception != null)
            {
                try
                {
                    _GetResponseDone.Set();
                }
                catch (ObjectDisposedException) { }
            }
        }
		/// SMTPコマンドの種類に基づいて状態を変化させます。
        /// <summary>
        /// SMTPコマンドの種類に基づいて状態を変化させます。
        /// </summary>
        /// <param name="command"></param>
        private void SetSmtpCommandState(SmtpCommand command)
        {
            if (command is Command.Mail)
            {
                this._State = SmtpConnectionState.MailFromCommandExecuting;
            }
            else if (command is Command.Rcpt)
            {
                this._State = SmtpConnectionState.RcptToCommandExecuting;
            }
            else if (command is Command.Data)
            {
                this._State = SmtpConnectionState.DataCommandExecuting;
            }
        }
        /// サーバーからのレスポンスの受信時に現在の状態に基づいて状態を変化させます。
        /// <summary>
        /// サーバーからのレスポンスの受信時に現在の状態に基づいて状態を変化させます。
        /// </summary>
        private void SetSmtpCommandState()
        {
            this._Commnicating = false;
            switch (this._State)
            {
                case SmtpConnectionState.MailFromCommandExecuting: this._State = SmtpConnectionState.MailFromCommandExecuted; break;
                case SmtpConnectionState.RcptToCommandExecuted: this._State = SmtpConnectionState.RcptToCommandExecuted; break;
                case SmtpConnectionState.DataCommandExecuting: this._State = SmtpConnectionState.DataCommandExecuted; break;
            }
        }
        /// SMTPサーバーに認証が必要かどうかを示す値を取得します。
        /// <summary>
        /// SMTPサーバーに認証が必要かどうかを示す値を取得します。
        /// </summary>
        /// <param name="text"></param>
        /// <returns></returns>
        private static Boolean NeedAuthenticate(String text)
        {
            return text.IndexOf("auth", StringComparison.InvariantCultureIgnoreCase) > -1;
        }
        /// StartTLSコマンドをサーバーに対して送信し、暗号化された通信を開始します。
        /// <summary>
        /// StartTLSコマンドをサーバーに対して送信し、暗号化された通信を開始します。
        /// </summary>
        private void StartTls()
        {
            SmtpCommandResult rs = null;

            if (this.EnsureOpen() == SmtpConnectionState.Connected)
            {
                rs = this.Execute("STARTTLS");
                if (rs.StatusCode != SmtpCommandResultCode.ServiceReady)
                { return; }

                this._Ssl = true;
                this._Tls = true;
                SslStream ssl = new SslStream(this._TcpClient.GetStream(), true, SmtpClient.RemoteCertificateValidationCallback, null);
                ssl.AuthenticateAsClient(this._ServerName);
                this._Stream = ssl;
            }
        }
        /// SMTPメールサーバーへログインします。
        /// <summary>
        /// SMTPメールサーバーへログインします。
        /// </summary>
        /// <returns></returns>
        public Boolean Authenticate()
        {
            SmtpCommandResult rs = null;

            if (this._PopBeforeSmtp == true)
            {
                Boolean bl = this._Pop3Client.Authenticate();
                this._Pop3Client.Close();
                if (bl == false)
                { return false; }
            }
            if (this._Mode == SmtpAuthenticateMode.Auto)
            {
                if (this.EnsureOpen() == SmtpConnectionState.Connected)
                {
                    rs = this.ExecuteEhlo();
                    String s = rs.Message.ToUpper();
                    //SMTP認証に対応している場合
                    if (s.Contains("AUTH") == true)
                    {
                        if (s.Contains("PLAIN") == true)
                        { return this.AuthenticateByPlain(); }
                        if (s.Contains("LOGIN") == true)
                        { return this.AuthenticateByLogin(); }
                        if (s.Contains("CRAM-MD5") == true)
                        { return this.AuthenticateByCramMD5(); }
                    }
                    else
                    {
                        rs = this.ExecuteEhlo();
                        return rs.StatusCode == SmtpCommandResultCode.ServiceReady;
                    }
                    //TLS認証
                    if (this.Tls == true)
                    {
                        if (s.Contains("STARTTLS") == false)
                        { throw new SmtpConnectException("TLS is not allowed."); }
                        this.StartTls();
                        rs = this.ExecuteEhlo();
                        return rs.StatusCode == SmtpCommandResultCode.ServiceReady;
                    }
                }
            }
            else
            {
                switch (this._Mode)
                {
                    case SmtpAuthenticateMode.None: return true;
                    case SmtpAuthenticateMode.Plain: return this.AuthenticateByPlain();
                    case SmtpAuthenticateMode.Login: return this.AuthenticateByLogin();
                    case SmtpAuthenticateMode.Cram_MD5: return this.AuthenticateByCramMD5();
                }
            }
            return false;
        }
        /// SMTPメールサーバーへPlain認証でログインします。
        /// <summary>
        /// SMTPメールサーバーへPlain認証でログインします。
        /// </summary>
        /// <returns></returns>
        public Boolean AuthenticateByPlain()
        {
            SmtpCommandResult rs = null;

            if (this.EnsureOpen() == SmtpConnectionState.Connected)
            {
                rs = this.Execute("Auth Plain");
                if (rs.StatusCode != SmtpCommandResultCode.WaitingForAuthentication)
                { return false; }
                //ユーザー名＆パスワードの文字列を送信
                rs = this.Execute(MailParser.ToBase64String(String.Format("{0}\0{0}\0{1}", this._UserName, this._Password)));
                if (rs.StatusCode == SmtpCommandResultCode.AuthenticationSuccessful)
                {
                    this._State = SmtpConnectionState.Authenticated;
                }
            }
            return this._State == SmtpConnectionState.Authenticated;
        }
        /// SMTPメールサーバーへLogin認証でログインします。
        /// <summary>
        /// SMTPメールサーバーへLogin認証でログインします。
        /// </summary>
        /// <returns></returns>
        public Boolean AuthenticateByLogin()
        {
            SmtpCommandResult rs = null;

            if (this.EnsureOpen() == SmtpConnectionState.Connected)
            {
                rs = this.Execute("Auth Login");
                if (rs.StatusCode != SmtpCommandResultCode.WaitingForAuthentication)
                { return false; }
                //ユーザー名送信
                rs = this.Execute(MailParser.ToBase64String(this._UserName));
                if (rs.StatusCode != SmtpCommandResultCode.WaitingForAuthentication)
                { return false; }
                //パスワード送信
                rs = this.Execute(MailParser.ToBase64String(this._Password));
                if (rs.StatusCode == SmtpCommandResultCode.AuthenticationSuccessful)
                {
                    this._State = SmtpConnectionState.Authenticated;
                }
            }
            return this._State == SmtpConnectionState.Authenticated;
        }
        /// SMTPメールサーバーへCRAM-MD5認証でログインします。
        /// <summary>
        /// SMTPメールサーバーへCRAM-MD5認証でログインします。
        /// </summary>
        /// <returns></returns>
        public Boolean AuthenticateByCramMD5()
        {
            SmtpCommandResult rs = null;

            if (this.EnsureOpen() == SmtpConnectionState.Connected)
            {
                rs = this.Execute("Auth CRAM-MD5");
                if (rs.StatusCode != SmtpCommandResultCode.WaitingForAuthentication)
                { return false; }
                //ユーザー名＋チャレンジ文字列をBase64エンコードした文字列を送信
                String s = MailParser.ToCramMd5String(rs.Message, this._UserName, this._Password);
                rs = this.Execute(s);
                if (rs.StatusCode == SmtpCommandResultCode.AuthenticationSuccessful)
                {
                    this._State = SmtpConnectionState.Authenticated;
                }
            }
            return this._State == SmtpConnectionState.Authenticated;
        }
        /// 同期でSMTPメールサーバーへコマンドを送信し、コマンドの種類によってはレスポンスデータを受信して返します。
        /// <summary>
        /// 同期でSMTPメールサーバーへコマンドを送信し、コマンドの種類によってはレスポンスデータを受信して返します。
        /// </summary>
        /// <param name="command"></param>
        /// <returns></returns>
        public SmtpCommandResult Execute(SmtpCommand command)
        {
            return this.Execute(command.GetCommandString());
        }
        /// 同期でSMTPメールサーバーへコマンドを送信し、コマンドの種類によってはレスポンスデータを受信して返します。
        /// <summary>
        /// 同期でSMTPメールサーバーへコマンドを送信し、コマンドの種類によってはレスポンスデータを受信して返します。
        /// </summary>
        /// <param name="command"></param>
        /// <returns></returns>
        private SmtpCommandResult Execute(String command)
        {
            this.SendCommand(command);
            this._Commnicating = true;
            return this.GetResponse();
        }
        /// SMTPメールサーバーへEHLOコマンドを送信します。
        /// <summary>
        /// SMTPメールサーバーへEHLOコマンドを送信します。
        /// </summary>
        /// <returns></returns>
        public SmtpCommandResult ExecuteEhlo()
        {
            this.EnsureOpen();
            return this.Execute(new Ehlo(this._HostName));
        }
        /// SMTPメールサーバーへHELOコマンドを送信します。
        /// <summary>
        /// SMTPメールサーバーへHELOコマンドを送信します。
        /// </summary>
        /// <returns></returns>
        public SmtpCommandResult ExecuteHelo()
        {
            this.EnsureOpen();
            return this.Execute(new Helo(this._HostName));
        }
        private SmtpCommandResult ExecuteEhloAndHelo()
        {
            SmtpCommandResult rs = null;
            //サーバーへメールトランザクションの開始コマンドを送信
            rs = this.ExecuteEhlo();
            if (rs.StatusCode != SmtpCommandResultCode.RequestedMailActionOkay_Completed)
            {
                rs = this.ExecuteHelo();
            }
            return rs;
        }
        /// SMTPメールサーバーへMAILコマンドを送信します。
        /// <summary>
        /// SMTPメールサーバーへMAILコマンドを送信します。
        /// </summary>
        /// <returns></returns>
        public SmtpCommandResult ExecuteMail(String reversePath)
        {
            this.EnsureOpen();
            return this.Execute(new Smtp.Command.Mail(reversePath));
        }
        /// SMTPメールサーバーへRCPTコマンドを送信します。
        /// <summary>
        /// SMTPメールサーバーへRCPTコマンドを送信します。
        /// </summary>
        /// <returns></returns>
        public SmtpCommandResult ExecuteRcpt(String forwardPath)
        {
            this.EnsureOpen();
            return this.Execute(new Rcpt(forwardPath));
        }
        /// SMTPメールサーバーへDATAコマンドを送信します。
        /// <summary>
        /// SMTPメールサーバーへDATAコマンドを送信します。
        /// </summary>
        /// <returns></returns>
        public SmtpCommandResult ExecuteData()
        {
            this.EnsureOpen();
            return this.Execute(new Data());
        }
        /// SMTPメールサーバーへRESETコマンドを送信します。
        /// <summary>
        /// SMTPメールサーバーへRESETコマンドを送信します。
        /// </summary>
        /// <returns></returns>
        public SmtpCommandResult ExecuteRset()
        {
            this.EnsureOpen();
            return this.Execute(new Rset());
        }
        /// SMTPメールサーバーへVRFYコマンドを送信します。
        /// <summary>
        /// SMTPメールサーバーへVRFYコマンドを送信します。
        /// </summary>
        /// <returns></returns>
        public SmtpCommandResult ExecuteVrfy(String userName)
        {
            this.EnsureOpen();
            return this.Execute(new Vrfy(userName));
        }
        /// SMTPメールサーバーへEXPNコマンドを送信します。
        /// <summary>
        /// SMTPメールサーバーへEXPNコマンドを送信します。
        /// </summary>
        /// <returns></returns>
        public SmtpCommandResult ExecuteExpn(String mailingList)
        {
            this.EnsureOpen();
            return this.Execute(new Expn(mailingList));
        }
        /// SMTPメールサーバーへHELPコマンドを送信します。
        /// <summary>
        /// SMTPメールサーバーへHELPコマンドを送信します。
        /// </summary>
        /// <returns></returns>
        public SmtpCommandResult ExecuteHelp()
        {
            this.EnsureOpen();
            return this.Execute(new Help());
        }
        /// SMTPメールサーバーへNOOPコマンドを送信します。
        /// <summary>
        /// SMTPメールサーバーへNOOPコマンドを送信します。
        /// </summary>
        /// <returns></returns>
        public SmtpCommandResult ExecuteNoop()
        {
            this.EnsureOpen();
            return this.Execute(new Noop());
        }
        /// SMTPメールサーバーへQUITコマンドを送信します。
        /// <summary>
        /// SMTPメールサーバーへQUITコマンドを送信します。
        /// </summary>
        /// <returns></returns>
        public SmtpCommandResult ExecuteQuit()
        {
            this.EnsureOpen();
            return this.Execute(new Quit());
        }
        /// メールを送信し、送信結果となるSendMailResultを取得します。
        /// <summary>
        /// メールを送信し、送信結果となるSendMailResultを取得します。
        /// </summary>
        /// <param name="message"></param>
        /// <returns></returns>
        public SendMailResult SendMail(SmtpMessage message)
        {
            if (message.From.StartsWith("<") == true &&
                message.From.EndsWith(">") == true)
            {
                return this.SendMail(message.From, message);
            }
            return this.SendMail("<" + message.From + ">", message);
        }
        /// メールを送信し、送信結果となるSendMailResultを取得します。
        /// <summary>
        /// メールを送信し、送信結果となるSendMailResultを取得します。
        /// </summary>
        /// <param name="from"></param>
        /// <param name="to"></param>
        /// <param name="cc"></param>
        /// <param name="bcc"></param>
        /// <param name="text"></param>
        /// <returns></returns>
        public SendMailResult SendMail(String from, String to, String cc, String bcc, String text)
        {
            List<MailAddress> l = new List<MailAddress>();
            String[] ss = null;

            ss = to.Split(',');
            for (int i = 0; i < ss.Length; i++)
            {
                if (String.IsNullOrEmpty(ss[i]) == true)
                { continue; }
                l.Add(MailAddress.Create(ss[i]));
            }
            ss = cc.Split(',');
            for (int i = 0; i < ss.Length; i++)
            {
                if (String.IsNullOrEmpty(ss[i]) == true)
                { continue; }
                l.Add(MailAddress.Create(ss[i]));
            }
            ss = bcc.Split(',');
            for (int i = 0; i < ss.Length; i++)
            {
                if (String.IsNullOrEmpty(ss[i]) == true)
                { continue; }
                l.Add(MailAddress.Create(ss[i]));
            }
            return this.SendMail(from, l, text);
        }
        /// メールを送信し、送信結果となるSendMailResultを取得します。
        /// <summary>
        /// メールを送信し、送信結果となるSendMailResultを取得します。
        /// </summary>
        /// <param name="from"></param>
        /// <param name="message"></param>
        /// <returns></returns>
        public SendMailResult SendMail(String from, SmtpMessage message)
        {
            List<MailAddress> l = new List<MailAddress>();
            l.AddRange(message.To);
            l.AddRange(message.Cc);
            l.AddRange(message.Bcc);
            return this.SendMail(from, l, message.GetDataText());
        }
        /// メールを送信し、送信結果となるSendMailResultを取得します。
        /// <summary>
        /// メールを送信し、送信結果となるSendMailResultを取得します。
        /// </summary>
        /// <param name="from"></param>
        /// <param name="rcptTo"></param>
        /// <param name="text"></param>
        /// <returns></returns>
        public SendMailResult SendMail(String from, List<MailAddress> rcptTo, String text)
        {
            SendMailResult r = new SendMailResult(false, text);
            SmtpCommandResult rs = null;
            Boolean HasRcpt = false;

            if (this.EnsureOpen() == SmtpConnectionState.Connected)
            {
                //サーバーへメールトランザクションの開始コマンドを送信
                rs = this.ExecuteEhloAndHelo();
                if (rs.StatusCode != SmtpCommandResultCode.RequestedMailActionOkay_Completed)
                { return r; }
                //TLS/SSL通信
                if (this._Tls == true)
                {
                    this.StartTls();
                    rs = this.ExecuteEhloAndHelo();
                }
                //ログイン認証が必要とされるかチェック
                if (SmtpClient.NeedAuthenticate(rs.Message) == true)
                {
                    if (this.Authenticate() == false)
                    { return r; }
                }
                //Mail Fromの送信
                rs = this.ExecuteMail(from);
                if (rs.StatusCode != SmtpCommandResultCode.RequestedMailActionOkay_Completed)
                { return r; }
                //Rcpt Toの送信
                foreach (var m in rcptTo)
                {
                    rs = this.ExecuteRcpt(m.ToString());
                    if (rs.StatusCode == SmtpCommandResultCode.RequestedMailActionOkay_Completed)
                    {
                        HasRcpt = true;
                    }
                    else
                    {
                        r.InvalidMailAddressList.Add(m);
                    }
                }
                if (HasRcpt == false)
                { return r; }
                //Dataの送信
                rs = this.ExecuteData();
                if (rs.StatusCode == SmtpCommandResultCode.StartMailInput)
                {
                    this.SendCommand(text + MailParser.NewLine + ".");
					rs = this.GetResponse();
                }
                else
                {
                    return r;
                }
                rs = this.ExecuteQuit();

                r.SendSuccessful = true;
                return r;
            }
            return new SendMailResult(false, ""); 
        }
        /// 終了処理を実行し、システムリソースを解放します。
        /// <summary>
        /// 終了処理を実行し、システムリソースを解放します。
        /// </summary>
        public void Dispose()
        {
            GC.SuppressFinalize(this);
            this.Dispose(true);
        }
		/// <summary>
		/// 
		/// </summary>
		/// <param name="disposing"></param>
        protected void Dispose(Boolean disposing)
        {
            if (disposing)
            { }
            if (this._TcpClient != null)
            {
                ((IDisposable)this._TcpClient).Dispose();
                this._TcpClient = null;
            }
            if (this._Pop3Client != null)
            {
                this._Pop3Client.Dispose();
            }
        }
		/// <summary>
		/// 
		/// </summary>
        ~SmtpClient()
        {
            this.Dispose(false);
        }
    }
}
