﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Net;
using System.Net.Sockets;
using Higuchi.Net.Mail;

namespace Higuchi.Net.Pop3
{
	public class Pop3Processor
	{
		public event EventHandler<Pop3AuthenticatedEventArgs> Authenticated;
		public event EventHandler<SocketMessageEventArgs> ReceiveMessage;
		public event EventHandler<SocketMessageEventArgs> SendMessage;
		public event ErrorEventHandler Error;
		private Socket _Socket = null;
		private Pop3Server _Pop3Server = null;
		private Pop3ProcessState _State = null;
		private TimeSpan _Timeout = new TimeSpan(0, 1, 0);
		public IPAddress DnsServerAddress
		{
			get { return _Pop3Server.DnsServerAddress; }
		}
		public IMailDataProvider DataProvider
		{
			get { return _Pop3Server.DataProvider; }
		}
		public Pop3Processor(Pop3Server smtpServer, Socket socket)
		{
			if (smtpServer == null) { throw new ArgumentNullException("smtpServer"); }
			if (socket == null) { throw new ArgumentNullException("socket"); }
			_Pop3Server = smtpServer;
			_Socket = socket;
		}
		public void Start()
		{
			String clientMessage = String.Empty;
			String serverMessage = String.Empty;
			NetworkStream ns = null;
			StreamReader sr = null;
			StreamWriter sw = null;
			DateTimeOffset previousTime = DateTimeOffset.Now;
			ServerReadyState ss= null;
			AuthenticatedState au = null;

			try
			{
				ns = new NetworkStream(_Socket);
				sr = new StreamReader(ns, Encoding.ASCII);
				sw = new StreamWriter(ns, Encoding.ASCII);
				sw.AutoFlush = true;

				ss = new ServerReadyState(this);
				_State = ss;
				sw.WriteLine("+OK server is ready for receiving data. <{0}>", ss.TimeStamp);
				while (true)
				{
					//Receive client message
					clientMessage = sr.ReadLine();

					au = _State as AuthenticatedState;
					_State = _State.ReceiveClientMessage(clientMessage);
					//Previous state is not AuthenticatedState and....
					if (au == null)
					{
						au = _State as AuthenticatedState;
						//This state is AuthenticatedState,raise authenticated event.
						if (au != null)
						{
							_Pop3Server.DataProvider.AuthenticatedUserNameList.Add(au.UserName);
							this.OnAuthenticated(au.UserName);
						}
					}
					//Send message to client
					using (StringReader r = new StringReader(_State.ServerMessage))
					{
						while (r.Peek() > -1)
						{
							sw.WriteLine(r.ReadLine());
						}
					}
					if (_State is QuitState) { break; }
					if (DateTimeOffset.Now > previousTime + _Timeout) { break; }
					previousTime = DateTimeOffset.Now;
				}
			}
			catch(Exception ex)
			{
				this.OnError(ex);
			}
			finally
			{
				if (sw != null) { sw.Close(); }
				if (sr != null) { sr.Close(); }
				if (ns != null) { ns.Close(); }
				au = _State as AuthenticatedState;
				if (au != null)
				{
					_Pop3Server.DataProvider.Quit(au.UserName);
					_Pop3Server.DataProvider.AuthenticatedUserNameList.Remove(au.UserName);
				}
				_Socket.Shutdown(SocketShutdown.Both);
				_Socket.Close();
			}
		}
		protected void OnAuthenticated(String userName)
		{
			if (this.Authenticated != null)
			{
				this.Authenticated(this, new Pop3AuthenticatedEventArgs(userName));
			}
		}
		protected void OnReceiveMessage(String text)
		{
			if (this.ReceiveMessage != null)
			{
				this.ReceiveMessage(this, new SocketMessageEventArgs(text));
			}
		}
		protected void OnSendMessage(String text)
		{
			if (this.SendMessage != null)
			{
				this.SendMessage(this, new SocketMessageEventArgs(text));
			}
		}
		protected void OnError(Exception ex)
		{
			if (this.Error != null)
			{
				this.Error(this, new ErrorEventArgs(ex));
			}
		}
	}
}
