﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Net;
using System.Net.Sockets;
using Higuchi.Net.Mail;

namespace Higuchi.Net.Smtp
{
	public class SmtpProcessor
	{
		public event EventHandler<SocketMessageEventArgs> ReceiveMessage;
		public event EventHandler<SocketMessageEventArgs> SendMessage;
		public event ErrorEventHandler Error;
		private Socket _Socket = null;
		private SmtpServer _SmtpServer = null;
		private SmtpProcessState _State = null;
		private TimeSpan _Timeout = new TimeSpan(0, 1, 0);
		public IPAddress DnsServerAddress
		{
			get { return _SmtpServer.DnsServerAddress; }
		}
		public IMailDataProvider DataProvider
		{
			get { return _SmtpServer.DataProvider; }
		}
		public SmtpProcessor(SmtpServer smtpServer, Socket socket)
		{
			if (smtpServer == null) { throw new ArgumentNullException("smtpServer"); }
			if (socket == null) { throw new ArgumentNullException("socket"); }
			_SmtpServer = smtpServer;
			_Socket = socket;
		}
		public void Start()
		{
			String clientMessage = String.Empty;
			String serverMessage = String.Empty;
			NetworkStream ns = null;
			StreamReader sr = null;
			StreamWriter sw = null;
			DateTimeOffset previousTime = DateTimeOffset.Now;
			
			try
			{
				ns = new NetworkStream(_Socket);
				sr = new StreamReader(ns, Encoding.ASCII);
				sw = new StreamWriter(ns, Encoding.ASCII);
				sw.AutoFlush = true;
				sw.WriteLine(SmtpServer.CreateMessage(SmtpCommandResultCode.ServiceReady));

				_State = new ServerReadyState(this);
				while (true)
				{
					//Receive client message
					clientMessage = sr.ReadLine();

					_State = _State.ReceiveClientMessage(clientMessage);
					//Send message to client
					using (StringReader r = new StringReader(_State.ServerMessage))
					{
						while (r.Peek() > -1)
						{
							sw.WriteLine(r.ReadLine());
						}
					}
					if (_State is QuitState) { break; }
					if (DateTimeOffset.Now > previousTime + _Timeout) { break; }
					previousTime = DateTimeOffset.Now;
				}
			}
			catch(Exception ex)
			{
				this.OnError(ex);
			}
			finally
			{
				if (sw != null) { sw.Close(); }
				if (sr != null) { sr.Close(); }
				if (ns != null) { ns.Close(); }
				_Socket.Shutdown(SocketShutdown.Both);
				_Socket.Close();
			}
		}
		public Boolean IsDelivery(MailAddress rcptTo)
		{
			return rcptTo.DomainName == _SmtpServer.DomainName;
		}
		public Boolean CanRelay(Boolean authenticated)
		{
			return _SmtpServer.RelayValidator.Validate(_Socket, authenticated);
		}
		protected void OnReceiveMessage(String text)
		{
			if (this.ReceiveMessage != null)
			{
				this.ReceiveMessage(this, new SocketMessageEventArgs(text));
			}
		}
		protected void OnSendMessage(String text)
		{
			if (this.SendMessage != null)
			{
				this.SendMessage(this, new SocketMessageEventArgs(text));
			}
		}
		protected void OnError(Exception ex)
		{
			if (this.Error != null)
			{
				this.Error(this, new ErrorEventArgs(ex));
			}
		}
	}
}
