﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using Higuchi.Net.Mail;

namespace Higuchi.Net.Smtp
{
	public class SmtpServer: Server
	{
		private IMailDataProvider _DataProvider = new NullDataProvider();
		private IRelayValidator _RelayValidator = new DefaultRelayValidator();
		private String _DomainName = "";
		public IMailDataProvider DataProvider
		{
			get { return _DataProvider; }
			set { _DataProvider = value; }
		}
		public IRelayValidator RelayValidator
		{
			get { return _RelayValidator; }
			set { _RelayValidator = value; }
		}
		public String DomainName
		{
			get { return _DomainName; }
			set { _DomainName = value; }
		}
		public SmtpServer(IPAddress address, Int32 port, String domainName)
			: base(address, port)
		{
			_RelayValidator = new DefaultRelayValidator();
			_DomainName = domainName;
		}
		public SmtpServer(IPAddress address, Int32 port, String domainName, IMailDataProvider provider)
			: base(address, port)
		{
			_RelayValidator = new DefaultRelayValidator();
			_DomainName = domainName;
			_DataProvider = provider;
		}
		/// <summary>
		/// This method is called as separated thread.
		/// </summary>
		/// <param name="socket"></param>
		protected override void OnConnected(Socket socket)
		{
			SmtpProcessor p = new SmtpProcessor(this, socket);
			p.ReceiveMessage += (o, e) => this.OnOutputText(e.Text);
			p.SendMessage += (o, e) => this.OnOutputText(e.Text);
			p.Error += (o, e) => this.OnOutputText(e.GetException().ToString());
			p.Start();
		}
		public static SmtpRequestCommand GetRequestCommand(String text)
		{
			if (String.IsNullOrEmpty(text) == true) { return SmtpRequestCommand.Unkown; }
			String s = text.ToUpper();
			if (s == "HELO" || s.StartsWith("HELO ") == true) { return SmtpRequestCommand.Helo; }
			if (s == "EHLO" || s.StartsWith("EHLO ") == true) { return SmtpRequestCommand.Ehlo; }
			if (s == "EXPN" || s.StartsWith("EXPN ") == true) { return SmtpRequestCommand.Expn; }
			if (s == "HELP" || s.StartsWith("HELP ") == true) { return SmtpRequestCommand.Help; }
			if (s.StartsWith("AUTH ", StringComparison.InvariantCultureIgnoreCase) == true) { return SmtpRequestCommand.Auth; }
			if (s.StartsWith("MAIL FROM:", StringComparison.InvariantCultureIgnoreCase) == true) { return SmtpRequestCommand.Mail; }
			if (s == "DATA") { return SmtpRequestCommand.Data; }
			if (s == "NOOP") { return SmtpRequestCommand.Noop; }
			if (s == "QUIT") { return SmtpRequestCommand.Quit; }
			if (s.StartsWith("RCPT TO:", StringComparison.InvariantCultureIgnoreCase) == true) { return SmtpRequestCommand.Rcpt; }
			if (s == "RSET") { return SmtpRequestCommand.Rset; }
			if (s.StartsWith("VRFY ", StringComparison.InvariantCultureIgnoreCase) == true) { return SmtpRequestCommand.Vrfy; }
			return SmtpRequestCommand.Unkown;
		}
		public static String CreateMessage(SmtpCommandResultCode code)
		{
			switch (code)
			{
				case SmtpCommandResultCode.None: return "";
				case SmtpCommandResultCode.SystemStatus_OrSystemHelpReply: return "211 Reply system help";
				case SmtpCommandResultCode.HelpMessage: return "214 help message!";
				case SmtpCommandResultCode.ServiceReady: return String.Format("220 {0} Smtp service ready at {1}"
						, Environment.MachineName, DateTimeOffset.Now.ToString());
				case SmtpCommandResultCode.ServiceClosingTransmissionChannel: return "221 Connection closed. Good-bye friend!";
				case SmtpCommandResultCode.AuthenticationSuccessful: return "235 Authentication successful";
				case SmtpCommandResultCode.RequestedMailActionOkay_Completed:
					return String.Format("250 response message by {0}.Hello [{1}]"
						, Environment.MachineName, Dns.GetHostEntry(Environment.MachineName).AddressList[0]);
				case SmtpCommandResultCode.UserNotLocal_WillForwardTo: return "251 User does not exist at local mailbox.It will be forward to destination.";
				case SmtpCommandResultCode.CannotVerifyUser_ButWillAcceptMessageAndAttemptDelivery: return "252 Fail user verify.But message attempt to delivery.";
				case SmtpCommandResultCode.WaitingForAuthentication: return "334 waiting for authentication.";
				case SmtpCommandResultCode.StartMailInput: return "354 Ready for receiving data.Go ahead.";
				case SmtpCommandResultCode.ServiceNotAvailable_ClosingTransmissionChannel: return "421 Service is not available.So closing transmission channel.";
				case SmtpCommandResultCode.APasswordTransitionIsNeeded: return "432 User need to change password.";
				case SmtpCommandResultCode.RequestedMailActionNotTaken_MailboxUnavailable: return "450 Requested mail action does not taken.";
				case SmtpCommandResultCode.RequestedActionAborted_ErrorInProcessing: return "451 Error in processing.";
				case SmtpCommandResultCode.TemporaryAuthenticationFailure: return "454 temporary authentication failure.";
				case SmtpCommandResultCode.SyntaxError_CommandUnrecognized: return "500 syntax error command unrecognized.";
				case SmtpCommandResultCode.SyntaxErrorInParametersOrArguments: return "501 syntax error in parameters or arguments.";
				case SmtpCommandResultCode.CommandNotImplemented: return "502 This command does not implemented.";
				case SmtpCommandResultCode.BadSequenceOfCommands: return "503 Bad sequence of command.";
				case SmtpCommandResultCode.CommandParameterNotImplemented: return "504 Command parameter does not implemented.";
				case SmtpCommandResultCode.AuthenticationRequired: return "530 Authentication required.";
				case SmtpCommandResultCode.EncryptionRequiredForRequestedAuthenticationMechanism: return "538 Encryption required for requested authentication.";
				case SmtpCommandResultCode.RequestedActionNotTaken_MailboxUnavailable: return "550 Requested mail action does not taken.Mailbox unavailable.";
				case SmtpCommandResultCode.UserNotLocal_PleaseTry: return "551 User does not exist.Forward your mail to forward-path.";
				case SmtpCommandResultCode.RequestedMailActionAborted_ExceededStorageAllocation: return "552 Requested mail action aborted.Storage allocation over.";
				case SmtpCommandResultCode.RequestedActionNotTaken_MailboxNameNotAllowed: return "553 Requested action does not taken.Mailbox name does not allowed.";
				case SmtpCommandResultCode.TransactionFailed: return "554 Transaction failed.";
			}
			throw new InvalidOperationException();
		}
		public static String FromBase64String(String text)
		{
			try
			{
				return MailParser.FromBase64String(text);
			}
			catch { }
			return "";
		}
	}
}
